class StreamReader(TextReader,IDisposable):
 """
 Implements a System.IO.TextReader that reads characters from a byte stream in a particular encoding.

 

 StreamReader(stream: Stream)

 StreamReader(stream: Stream,detectEncodingFromByteOrderMarks: bool)

 StreamReader(stream: Stream,encoding: Encoding)

 StreamReader(stream: Stream,encoding: Encoding,detectEncodingFromByteOrderMarks: bool)

 StreamReader(stream: Stream,encoding: Encoding,detectEncodingFromByteOrderMarks: bool,bufferSize: int)

 StreamReader(stream: Stream,encoding: Encoding,detectEncodingFromByteOrderMarks: bool,bufferSize: int,leaveOpen: bool)

 StreamReader(path: str)

 StreamReader(path: str,detectEncodingFromByteOrderMarks: bool)

 StreamReader(path: str,encoding: Encoding)

 StreamReader(path: str,encoding: Encoding,detectEncodingFromByteOrderMarks: bool)

 StreamReader(path: str,encoding: Encoding,detectEncodingFromByteOrderMarks: bool,bufferSize: int)
 """
 def Close(self):
  """
  Close(self: StreamReader)

   Closes the System.IO.StreamReader object and the underlying stream,and releases any system 

    resources associated with the reader.
  """
  pass
 def DiscardBufferedData(self):
  """
  DiscardBufferedData(self: StreamReader)

   Clears the internal buffer.
  """
  pass
 def Dispose(self):
  """
  Dispose(self: StreamReader,disposing: bool)

   Closes the underlying stream,releases the unmanaged resources used by the 

    System.IO.StreamReader,and optionally releases the managed resources.

  

  

   disposing: true to release both managed and unmanaged resources; false to release only unmanaged resources.
  """
  pass
 def MemberwiseClone(self,*args):
  """
  MemberwiseClone(self: MarshalByRefObject,cloneIdentity: bool) -> MarshalByRefObject

  

   Creates a shallow copy of the current System.MarshalByRefObject object.

  

   cloneIdentity: false to delete the current System.MarshalByRefObject object's identity,which will cause the 

    object to be assigned a new identity when it is marshaled across a remoting boundary. A value of 

    false is usually appropriate. true to copy the current System.MarshalByRefObject object's 

    identity to its clone,which will cause remoting client calls to be routed to the remote server 

    object.

  

   Returns: A shallow copy of the current System.MarshalByRefObject object.

  MemberwiseClone(self: object) -> object

  

   Creates a shallow copy of the current System.Object.

   Returns: A shallow copy of the current System.Object.
  """
  pass
 def Peek(self):
  """
  Peek(self: StreamReader) -> int

  

   Returns the next available character but does not consume it.

   Returns: An integer representing the next character to be read,or -1 if there are no characters to be 

    read or if the stream does not support seeking.
  """
  pass
 def Read(self,buffer=None,index=None,count=None):
  """
  Read(self: StreamReader,index: int,count: int) -> (int,Array[Char])

  

   Reads a specified maximum of characters from the current stream into a buffer,beginning at the 

    specified index.

  

  

   index: The index of buffer at which to begin writing.

   count: The maximum number of characters to read.

   Returns: The number of characters that have been read,or 0 if at the end of the stream and no data was 

    read. The number will be less than or equal to the count parameter,depending on whether the 

    data is available within the stream.

  

  Read(self: StreamReader) -> int

  

   Reads the next character from the input stream and advances the character position by one 

    character.

  

   Returns: The next character from the input stream represented as an System.Int32 object,or -1 if no more 

    characters are available.
  """
  pass
 def ReadAsync(self,buffer,index,count):
  """ ReadAsync(self: StreamReader,buffer: Array[Char],index: int,count: int) -> Task[int] """
  pass
 def ReadBlock(self,buffer,index,count):
  """ ReadBlock(self: StreamReader,index: int,count: int) -> (int,Array[Char]) """
  pass
 def ReadBlockAsync(self,buffer,index,count):
  """ ReadBlockAsync(self: StreamReader,buffer: Array[Char],index: int,count: int) -> Task[int] """
  pass
 def ReadLine(self):
  """
  ReadLine(self: StreamReader) -> str

  

   Reads a line of characters from the current stream and returns the data as a string.

   Returns: The next line from the input stream,or null if the end of the input stream is reached.
  """
  pass
 def ReadLineAsync(self):
  """ ReadLineAsync(self: StreamReader) -> Task[str] """
  pass
 def ReadToEnd(self):
  """
  ReadToEnd(self: StreamReader) -> str

  

   Reads the stream from the current position to the end of the stream.

   Returns: The rest of the stream as a string,from the current position to the end. If the current 

    position is at the end of the stream,returns an empty string ("").
  """
  pass
 def ReadToEndAsync(self):
  """ ReadToEndAsync(self: StreamReader) -> Task[str] """
  pass
 def __enter__(self,*args):
  """
  __enter__(self: IDisposable) -> object

  

   Provides the implementation of __enter__ for objects which implement IDisposable.
  """
  pass
 def __exit__(self,*args):
  """
  __exit__(self: IDisposable,exc_type: object,exc_value: object,exc_back: object)

   Provides the implementation of __exit__ for objects which implement IDisposable.
  """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 @staticmethod
 def __new__(self,*__args):
  """
  __new__(cls: type,stream: Stream)

  __new__(cls: type,stream: Stream,detectEncodingFromByteOrderMarks: bool)

  __new__(cls: type,stream: Stream,encoding: Encoding)

  __new__(cls: type,stream: Stream,encoding: Encoding,detectEncodingFromByteOrderMarks: bool)

  __new__(cls: type,stream: Stream,encoding: Encoding,detectEncodingFromByteOrderMarks: bool,bufferSize: int)

  __new__(cls: type,stream: Stream,encoding: Encoding,detectEncodingFromByteOrderMarks: bool,bufferSize: int,leaveOpen: bool)

  __new__(cls: type,path: str)

  __new__(cls: type,path: str,detectEncodingFromByteOrderMarks: bool)

  __new__(cls: type,path: str,encoding: Encoding)

  __new__(cls: type,path: str,encoding: Encoding,detectEncodingFromByteOrderMarks: bool)

  __new__(cls: type,path: str,encoding: Encoding,detectEncodingFromByteOrderMarks: bool,bufferSize: int)
  """
  pass
 def __reduce_ex__(self,*args):
  pass
 BaseStream=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Returns the underlying stream.



Get: BaseStream(self: StreamReader) -> Stream



"""

 CurrentEncoding=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the current character encoding that the current System.IO.StreamReader object is using.



Get: CurrentEncoding(self: StreamReader) -> Encoding



"""

 EndOfStream=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value that indicates whether the current stream position is at the end of the stream.



Get: EndOfStream(self: StreamReader) -> bool



"""


 Null=None




class DrawListViewItemEventArgs(EventArgs):
 """
 Provides data for the System.Windows.Forms.ListView.DrawItem event.

 

 DrawListViewItemEventArgs(graphics: Graphics,item: ListViewItem,bounds: Rectangle,itemIndex: int,state: ListViewItemStates)
 """
 def DrawBackground(self):
  """
  DrawBackground(self: DrawListViewItemEventArgs)

   Draws the background of the System.Windows.Forms.ListViewItem using its current background color.
  """
  pass
 def DrawFocusRectangle(self):
  """
  DrawFocusRectangle(self: DrawListViewItemEventArgs)

   Draws a focus rectangle for the System.Windows.Forms.ListViewItem if it has focus.
  """
  pass
 def DrawText(self,flags=None):
  """
  DrawText(self: DrawListViewItemEventArgs,flags: TextFormatFlags)

   Draws the text of the System.Windows.Forms.ListViewItem using its current foreground color and 

    formatting it with the specified System.Windows.Forms.TextFormatFlags values.

  

  

   flags: A bitwise combination of System.Windows.Forms.TextFormatFlags values.

  DrawText(self: DrawListViewItemEventArgs)

   Draws the text of the System.Windows.Forms.ListViewItem using its current foreground color.
  """
  pass
 def __getitem__(self,*args):
  """ x.__getitem__(y) <==> x[y] """
  pass
 @staticmethod
 def __new__(self,graphics,item,bounds,itemIndex,state):
  """ __new__(cls: type,graphics: Graphics,item: ListViewItem,bounds: Rectangle,itemIndex: int,state: ListViewItemStates) """
  pass
 Bounds=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the size and location of the System.Windows.Forms.ListViewItem to draw.



Get: Bounds(self: DrawListViewItemEventArgs) -> Rectangle



"""

 DrawDefault=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets a property indicating whether the System.Windows.Forms.ListView control will use the default drawing for the System.Windows.Forms.ListViewItem.



Get: DrawDefault(self: DrawListViewItemEventArgs) -> bool



Set: DrawDefault(self: DrawListViewItemEventArgs)=value

"""

 Graphics=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the System.Drawing.Graphics used to draw the System.Windows.Forms.ListViewItem.



Get: Graphics(self: DrawListViewItemEventArgs) -> Graphics



"""

 Item=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the System.Windows.Forms.ListViewItem to draw.



Get: Item(self: DrawListViewItemEventArgs) -> ListViewItem



"""

 ItemIndex=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the index of the System.Windows.Forms.ListViewItem within the System.Windows.Forms.ListView.Items collection of the containing System.Windows.Forms.ListView.



Get: ItemIndex(self: DrawListViewItemEventArgs) -> int



"""

 State=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the current state of the System.Windows.Forms.ListViewItem to draw.



Get: State(self: DrawListViewItemEventArgs) -> ListViewItemStates



"""





class ConsoleKey(Enum,IComparable,IFormattable,IConvertible):
 """
 Specifies the standard keys on a console.

 

 enum ConsoleKey,values: A (65),Add (107),Applications (93),Attention (246),B (66),Backspace (8),BrowserBack (166),BrowserFavorites (171),BrowserForward (167),BrowserHome (172),BrowserRefresh (168),BrowserSearch (170),BrowserStop (169),C (67),Clear (12),CrSel (247),D (68),D0 (48),D1 (49),D2 (50),D3 (51),D4 (52),D5 (53),D6 (54),D7 (55),D8 (56),D9 (57),Decimal (110),Delete (46),Divide (111),DownArrow (40),E (69),End (35),Enter (13),EraseEndOfFile (249),Escape (27),Execute (43),ExSel (248),F (70),F1 (112),F10 (121),F11 (122),F12 (123),F13 (124),F14 (125),F15 (126),F16 (127),F17 (128),F18 (129),F19 (130),F2 (113),F20 (131),F21 (132),F22 (133),F23 (134),F24 (135),F3 (114),F4 (115),F5 (116),F6 (117),F7 (118),F8 (119),F9 (120),G (71),H (72),Help (47),Home (36),I (73),Insert (45),J (74),K (75),L (76),LaunchApp1 (182),LaunchApp2 (183),LaunchMail (180),LaunchMediaSelect (181),LeftArrow (37),LeftWindows (91),M (77),MediaNext (176),MediaPlay (179),MediaPrevious (177),MediaStop (178),Multiply (106),N (78),NoName (252),NumPad0 (96),NumPad1 (97),NumPad2 (98),NumPad3 (99),NumPad4 (100),NumPad5 (101),NumPad6 (102),NumPad7 (103),NumPad8 (104),NumPad9 (105),O (79),Oem1 (186),Oem102 (226),Oem2 (191),Oem3 (192),Oem4 (219),Oem5 (220),Oem6 (221),Oem7 (222),Oem8 (223),OemClear (254),OemComma (188),OemMinus (189),OemPeriod (190),OemPlus (187),P (80),Pa1 (253),Packet (231),PageDown (34),PageUp (33),Pause (19),Play (250),Print (42),PrintScreen (44),Process (229),Q (81),R (82),RightArrow (39),RightWindows (92),S (83),Select (41),Separator (108),Sleep (95),Spacebar (32),Subtract (109),T (84),Tab (9),U (85),UpArrow (38),V (86),VolumeDown (174),VolumeMute (173),VolumeUp (175),W (87),X (88),Y (89),Z (90),Zoom (251)
 """
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==yx.__eq__(y) <==> x==yx.__eq__(y) <==> x==y """
  pass
 def __format__(self,*args):
  """ __format__(formattable: IFormattable,format: str) -> str """
  pass
 def __ge__(self,*args):
  pass
 def __gt__(self,*args):
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __le__(self,*args):
  pass
 def __lt__(self,*args):
  pass
 def __ne__(self,*args):
  pass
 def __reduce_ex__(self,*args):
  pass
 def __str__(self,*args):
  pass
 A=None
 Add=None
 Applications=None
 Attention=None
 B=None
 Backspace=None
 BrowserBack=None
 BrowserFavorites=None
 BrowserForward=None
 BrowserHome=None
 BrowserRefresh=None
 BrowserSearch=None
 BrowserStop=None
 C=None
 Clear=None
 CrSel=None
 D=None
 D0=None
 D1=None
 D2=None
 D3=None
 D4=None
 D5=None
 D6=None
 D7=None
 D8=None
 D9=None
 Decimal=None
 Delete=None
 Divide=None
 DownArrow=None
 E=None
 End=None
 Enter=None
 EraseEndOfFile=None
 Escape=None
 Execute=None
 ExSel=None
 F=None
 F1=None
 F10=None
 F11=None
 F12=None
 F13=None
 F14=None
 F15=None
 F16=None
 F17=None
 F18=None
 F19=None
 F2=None
 F20=None
 F21=None
 F22=None
 F23=None
 F24=None
 F3=None
 F4=None
 F5=None
 F6=None
 F7=None
 F8=None
 F9=None
 G=None
 H=None
 Help=None
 Home=None
 I=None
 Insert=None
 J=None
 K=None
 L=None
 LaunchApp1=None
 LaunchApp2=None
 LaunchMail=None
 LaunchMediaSelect=None
 LeftArrow=None
 LeftWindows=None
 M=None
 MediaNext=None
 MediaPlay=None
 MediaPrevious=None
 MediaStop=None
 Multiply=None
 N=None
 NoName=None
 NumPad0=None
 NumPad1=None
 NumPad2=None
 NumPad3=None
 NumPad4=None
 NumPad5=None
 NumPad6=None
 NumPad7=None
 NumPad8=None
 NumPad9=None
 O=None
 Oem1=None
 Oem102=None
 Oem2=None
 Oem3=None
 Oem4=None
 Oem5=None
 Oem6=None
 Oem7=None
 Oem8=None
 OemClear=None
 OemComma=None
 OemMinus=None
 OemPeriod=None
 OemPlus=None
 P=None
 Pa1=None
 Packet=None
 PageDown=None
 PageUp=None
 Pause=None
 Play=None
 Print=None
 PrintScreen=None
 Process=None
 Q=None
 R=None
 RightArrow=None
 RightWindows=None
 S=None
 Select=None
 Separator=None
 Sleep=None
 Spacebar=None
 Subtract=None
 T=None
 Tab=None
 U=None
 UpArrow=None
 V=None
 value__=None
 VolumeDown=None
 VolumeMute=None
 VolumeUp=None
 W=None
 X=None
 Y=None
 Z=None
 Zoom=None




# -*- coding: utf-8 -*-
"""Route dictionary."""
#pylint: disable=import-error,invalid-name,broad-except
import re
import uuid
from collections import namedtuple

from pyrevit.coreutils import envvars
from pyrevit.coreutils.logger import get_logger

from pyrevit.routes.server import handler


mlogger = get_logger(__name__)


ROUTE_VAR_SEP = ':'


Route = namedtuple('Route', ['pattern', 'method'])

RouteParam = namedtuple('RouteParam', ['key', 'value'])


def _make_finder_pattern(route_pattern):
    modified_pattern = re.sub(r"\<.+?\>", "(.+)", route_pattern)
    return '^%s$' % modified_pattern


def _find_pattern_keys(route_pattern):
    return re.findall(r"\<(.+?)\>", route_pattern)


def _find_pattern_values(finder_pattern, path):
    values = re.findall(finder_pattern, path)
    if values:
        # if more that one match, it returns a tuple
        if isinstance(values[0], tuple):
            return [x for x in values[0]]
        return values


def _extract_route_param_keys(route_pattern):
    clean_keys = set()
    for param_key in _find_pattern_keys(route_pattern):
        if ROUTE_VAR_SEP in param_key:
            clean_keys.add(param_key.split(ROUTE_VAR_SEP)[1])
        else:
            clean_keys.add(param_key)
    return clean_keys


def _validate_pattern(route_pattern):
    param_keys = _extract_route_param_keys(route_pattern)
    return not param_keys.intersection(handler.RESERVED_VAR_NAMES)


def _cast_value(cast, val):
    if cast in ['int', 'float', 'double', 'bool']:
        return eval('%s(%s)' % (cast, val)) #pylint: disable=eval-used
    elif cast == 'uuid':
        return uuid.UUID(val)


def route_match(route, path, method):
    """Test if route pattern matches given request path."""
    finder_pattern = _make_finder_pattern(route.pattern)
    # if same method and matching pattern
    if route.method == method \
            and re.match(finder_pattern, path):
        # check variable data types
        # e.g. paths below are different
        # api/v1/posts/<int:pid> matches
        #     api/v1/posts/12
        # api/v1/posts/<int:pid> does not match
        #     api/v1/posts/661a4f7a-7377-11ea-9494-acde48001122
        for key, val in zip(
                _find_pattern_keys(route.pattern),
                _find_pattern_values(finder_pattern, path)
            ):
            if ROUTE_VAR_SEP in key:
                cast, key = key.split(ROUTE_VAR_SEP)
                try:
                    val = _cast_value(cast, val)
                except Exception:
                    # if any of variable types does not match, say no
                    return False
        # if all variable types matched, say yes
        return True
    # otherwise say no
    return False


def extract_route_params(route_pattern, request_path):
    """Extracts route params from request path based on pattern.

    Examples:
        ```python
        extract_route_params('api/v1/posts/<int:id>', 'api/v1/posts/12')
        ```
        [<RouteParam key:id value=12>]
    """
    finder_pattern = _make_finder_pattern(route_pattern)
    route_params = []
    for key, val in zip(
            _find_pattern_keys(route_pattern),
            _find_pattern_values(finder_pattern, request_path)
        ):
        if ROUTE_VAR_SEP in key:
            cast, key = key.split(ROUTE_VAR_SEP)
            try:
                val = _cast_value(cast, val)
            except Exception as cast_ex:
                mlogger.debug(
                    'Cast error %s -> %s | %s', val, cast, str(cast_ex)
                    )
        route_params.append(
            RouteParam(key=key, value=val)
        )
    return route_params


def reset_routes():
    """Reset registered APIs and routes."""
    envvars.set_pyrevit_env_var(envvars.ROUTES_ROUTES, {})


def get_routes(api_name):
    """Get all registered routes for given API name.

    Args:
        api_name (str): unique name of the api

    Returns:
        (dict[str, Caller[]]): registered routes
    """
    if api_name is None:
        raise Exception("API name can not be None.")

    routes_map = envvars.get_pyrevit_env_var(envvars.ROUTES_ROUTES)
    if routes_map is None:
        routes_map = {}
        envvars.set_pyrevit_env_var(envvars.ROUTES_ROUTES, routes_map)

    if api_name in routes_map:
        return routes_map[api_name]
    else:
        app_routes = {}
        routes_map[api_name] = app_routes
        return app_routes


def get_route_handler(api_name, path, method):
    """Return registered handler for given API, path, and method.

    Args:
        api_name (str): unique name of the api
        path (str): request path
        method (str): method name

    Returns:
        api_route (Route): API route
        route_handler (Caller): registered route handler function
    """
    for api_route, route_handler in get_routes(api_name).items():
        if route_match(api_route, path, method):
            return api_route, route_handler
    return None, None


def add_route(api_name, pattern, method, handler_func):
    """Add new route for given API name.

    Args:
        api_name (str): unique name of the api
        pattern (str): route pattern
        method (str): method name
        handler_func (function): route handler function
    """
    if _validate_pattern(pattern):
        route = Route(pattern=pattern, method=method)
        app_routes = get_routes(api_name)
        app_routes[route] = handler_func
    else:
        mlogger.error('Route pattern is invalid: %s', pattern)


def remove_route(api_name, pattern, method):
    """Remove prevriously defined route for given API name.

    Args:
        api_name (str): unique name of the api
        pattern (str): route pattern
        method (str): method name
    """
    route = Route(pattern=pattern, method=method)
    app_routes = get_routes(api_name)
    app_routes.pop(route)



class INotifyPropertyChanging:
 """ Notifies clients that a property value is changing. """
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 PropertyChanging=None




class Version(object,ICloneable,IComparable,IComparable[Version],IEquatable[Version]):
 """
 Represents the version number of an assembly,operating system,or the common language runtime. This class cannot be inherited.

 

 Version(major: int,minor: int,build: int,revision: int)

 Version(major: int,minor: int,build: int)

 Version(major: int,minor: int)

 Version(version: str)

 Version()
 """
 def Clone(self):
  """
  Clone(self: Version) -> object

  

   Returns a new System.Version object whose value is the same as the current System.Version object.

   Returns: A new System.Object whose values are a copy of the current System.Version object.
  """
  pass
 def CompareTo(self,*__args):
  """
  CompareTo(self: Version,value: Version) -> int

  

   Compares the current System.Version object to a specified System.Version object and returns an 

    indication of their relative values.

  

  

   value: A System.Version object to compare to the current System.Version object,or null.

   Returns: A signed integer that indicates the relative values of the two objects,as shown in the 

    following table.Return value Meaning Less than zero The current System.Version object is a 

    version before value. Zero The current System.Version object is the same version as value. 

    Greater than zero The current System.Version object is a version subsequent to value. -or-value 

    is null.

  

  CompareTo(self: Version,version: object) -> int

  

   Compares the current System.Version object to a specified object and returns an indication of 

    their relative values.

  

  

   version: An object to compare,or null.

   Returns: A signed integer that indicates the relative values of the two objects,as shown in the 

    following table.Return value Meaning Less than zero The current System.Version object is a 

    version before version. Zero The current System.Version object is the same version as version. 

    Greater than zero The current System.Version object is a version subsequent to version.-or- 

    version is null.
  """
  pass
 def Equals(self,obj):
  """
  Equals(self: Version,obj: Version) -> bool

  

   Returns a value indicating whether the current System.Version object and a specified 

    System.Version object represent the same value.

  

  

   obj: A System.Version object to compare to the current System.Version object,or null.

   Returns: true if every component of the current System.Version object matches the corresponding component 

    of the obj parameter; otherwise,false.

  

  Equals(self: Version,obj: object) -> bool

  

   Returns a value indicating whether the current System.Version object is equal to a specified 

    object.

  

  

   obj: An object to compare with the current System.Version object,or null.

   Returns: true if the current System.Version object and obj are both System.Version objects,and every 

    component of the current System.Version object matches the corresponding component of obj; 

    otherwise,false.
  """
  pass
 def GetHashCode(self):
  """
  GetHashCode(self: Version) -> int

  

   Returns a hash code for the current System.Version object.

   Returns: A 32-bit signed integer hash code.
  """
  pass
 @staticmethod
 def Parse(input):
  """
  Parse(input: str) -> Version

  

   Converts the string representation of a version number to an equivalent System.Version object.

  

   input: A string that contains a version number to convert.

   Returns: An object that is equivalent to the version number specified in the input parameter.
  """
  pass
 def ToString(self,fieldCount=None):
  """
  ToString(self: Version,fieldCount: int) -> str

  

   Converts the value of the current System.Version object to its equivalent System.String 

    representation. A specified count indicates the number of components to return.

  

  

   fieldCount: The number of components to return. The fieldCount ranges from 0 to 4.

   Returns: The System.String representation of the values of the major,minor,build,and revision 

    components of the current System.Version object,each separated by a period character ('.'). The 

    fieldCount parameter determines how many components are returned.fieldCount Return Value 0 An 

    empty string (""). 1 major 2 major.minor 3 major.minor.build 4 major.minor.build.revision For 

    example,if you create System.Version object using the constructor Version(1,3,5),ToString(2) 

    returns "1.3" and ToString(4) throws an exception.

  

  ToString(self: Version) -> str

  

   Converts the value of the current System.Version object to its equivalent System.String 

    representation.

  

   Returns: The System.String representation of the values of the major,minor,build,and revision 

    components of the current System.Version object,as depicted in the following format. Each 

    component is separated by a period character ('.'). Square brackets ('[' and ']') indicate a 

    component that will not appear in the return value if the component is not defined: 

    major.minor[.build[.revision]] For example,if you create a System.Version object using the 

    constructor Version(1,1),the returned string is "1.1". If you create a System.Version object 

    using the constructor Version(1,3,4,2),the returned string is "1.3.4.2".
  """
  pass
 @staticmethod
 def TryParse(input,result):
  """
  TryParse(input: str) -> (bool,Version)

  

   Tries to convert the string representation of a version number to an equivalent System.Version 

    object,and returns a value that indicates whether the conversion succeeded.

  

  

   input: A string that contains a version number to convert.

   Returns: true if the input parameter was converted successfully; otherwise,false.
  """
  pass
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==y """
  pass
 def __ge__(self,*args):
  pass
 def __gt__(self,*args):
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __le__(self,*args):
  pass
 def __lt__(self,*args):
  pass
 @staticmethod
 def __new__(self,*__args):
  """
  __new__(cls: type,major: int,minor: int,build: int,revision: int)

  __new__(cls: type,major: int,minor: int,build: int)

  __new__(cls: type,major: int,minor: int)

  __new__(cls: type,version: str)

  __new__(cls: type)
  """
  pass
 def __ne__(self,*args):
  pass
 def __reduce_ex__(self,*args):
  pass
 def __repr__(self,*args):
  """ __repr__(self: object) -> str """
  pass
 def __str__(self,*args):
  pass
 Build=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the value of the build component of the version number for the current System.Version object.



Get: Build(self: Version) -> int



"""

 Major=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the value of the major component of the version number for the current System.Version object.



Get: Major(self: Version) -> int



"""

 MajorRevision=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the high 16 bits of the revision number.



Get: MajorRevision(self: Version) -> Int16



"""

 Minor=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the value of the minor component of the version number for the current System.Version object.



Get: Minor(self: Version) -> int



"""

 MinorRevision=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the low 16 bits of the revision number.



Get: MinorRevision(self: Version) -> Int16



"""

 Revision=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the value of the revision component of the version number for the current System.Version object.



Get: Revision(self: Version) -> int



"""





class ToolStripSystemRenderer(ToolStripRenderer):
 """
 Handles the painting functionality for System.Windows.Forms.ToolStrip objects,using system colors and a flat visual style.

 

 ToolStripSystemRenderer()
 """



class HttpListenerResponse(object,IDisposable):
 """ Represents a response to a request being handled by an System.Net.HttpListener object. """
 def Abort(self):
  """
  Abort(self: HttpListenerResponse)

   Closes the connection to the client without sending a response.
  """
  pass
 def AddHeader(self,name,value):
  """
  AddHeader(self: HttpListenerResponse,name: str,value: str)

   Adds the specified header and value to the HTTP headers for this response.

  

   name: The name of the HTTP header to set.

   value: The value for the name header.
  """
  pass
 def AppendCookie(self,cookie):
  """
  AppendCookie(self: HttpListenerResponse,cookie: Cookie)

   Adds the specified System.Net.Cookie to the collection of cookies for this response.

  

   cookie: The System.Net.Cookie to add to the collection to be sent with this response
  """
  pass
 def AppendHeader(self,name,value):
  """
  AppendHeader(self: HttpListenerResponse,name: str,value: str)

   Appends a value to the specified HTTP header to be sent with this response.

  

   name: The name of the HTTP header to append value to.

   value: The value to append to the name header.
  """
  pass
 def Close(self,responseEntity=None,willBlock=None):
  """
  Close(self: HttpListenerResponse)

   Sends the response to the client and releases the resources held by this 

    System.Net.HttpListenerResponse instance.

  

  Close(self: HttpListenerResponse,responseEntity: Array[Byte],willBlock: bool)

   Returns the specified byte array to the client and releases the resources held by this 

    System.Net.HttpListenerResponse instance.

  

  

   responseEntity: A System.Byte array that contains the response to send to the client.

   willBlock: true to block execution while flushing the stream to the client; otherwise,false.
  """
  pass
 def CopyFrom(self,templateResponse):
  """
  CopyFrom(self: HttpListenerResponse,templateResponse: HttpListenerResponse)

   Copies properties from the specified System.Net.HttpListenerResponse to this response.

  

   templateResponse: The System.Net.HttpListenerResponse instance to copy.
  """
  pass
 def Redirect(self,url):
  """
  Redirect(self: HttpListenerResponse,url: str)

   Configures the response to redirect the client to the specified URL.

  

   url: The URL that the client should use to locate the requested resource.
  """
  pass
 def SetCookie(self,cookie):
  """
  SetCookie(self: HttpListenerResponse,cookie: Cookie)

   Adds or updates a System.Net.Cookie in the collection of cookies sent with this response.

  

   cookie: A System.Net.Cookie for this response.
  """
  pass
 def __enter__(self,*args):
  """
  __enter__(self: IDisposable) -> object

  

   Provides the implementation of __enter__ for objects which implement IDisposable.
  """
  pass
 def __exit__(self,*args):
  """
  __exit__(self: IDisposable,exc_type: object,exc_value: object,exc_back: object)

   Provides the implementation of __exit__ for objects which implement IDisposable.
  """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __repr__(self,*args):
  """ __repr__(self: object) -> str """
  pass
 ContentEncoding=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the System.Text.Encoding for this response's System.Net.HttpListenerResponse.OutputStream.



Get: ContentEncoding(self: HttpListenerResponse) -> Encoding



Set: ContentEncoding(self: HttpListenerResponse)=value

"""

 ContentLength64=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the number of bytes in the body data included in the response.



Get: ContentLength64(self: HttpListenerResponse) -> Int64



Set: ContentLength64(self: HttpListenerResponse)=value

"""

 ContentType=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the MIME type of the content returned.



Get: ContentType(self: HttpListenerResponse) -> str



Set: ContentType(self: HttpListenerResponse)=value

"""

 Cookies=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the collection of cookies returned with the response.



Get: Cookies(self: HttpListenerResponse) -> CookieCollection



Set: Cookies(self: HttpListenerResponse)=value

"""

 Headers=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the collection of header name/value pairs returned by the server.



Get: Headers(self: HttpListenerResponse) -> WebHeaderCollection



Set: Headers(self: HttpListenerResponse)=value

"""

 KeepAlive=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets a value indicating whether the server requests a persistent connection.



Get: KeepAlive(self: HttpListenerResponse) -> bool



Set: KeepAlive(self: HttpListenerResponse)=value

"""

 OutputStream=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a System.IO.Stream object to which a response can be written.



Get: OutputStream(self: HttpListenerResponse) -> Stream



"""

 ProtocolVersion=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the HTTP version used for the response.



Get: ProtocolVersion(self: HttpListenerResponse) -> Version



Set: ProtocolVersion(self: HttpListenerResponse)=value

"""

 RedirectLocation=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the value of the HTTP Location header in this response.



Get: RedirectLocation(self: HttpListenerResponse) -> str



Set: RedirectLocation(self: HttpListenerResponse)=value

"""

 SendChunked=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets whether the response uses chunked transfer encoding.



Get: SendChunked(self: HttpListenerResponse) -> bool



Set: SendChunked(self: HttpListenerResponse)=value

"""

 StatusCode=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the HTTP status code to be returned to the client.



Get: StatusCode(self: HttpListenerResponse) -> int



Set: StatusCode(self: HttpListenerResponse)=value

"""

 StatusDescription=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets a text description of the HTTP status code returned to the client.



Get: StatusDescription(self: HttpListenerResponse) -> str



Set: StatusDescription(self: HttpListenerResponse)=value

"""





"""
Parameter Wrapper

>>> wrapped_element.parameters['Length']
10.0
>>> wrapped_element.parameters['Length'] = 5
5.0

"""  #
from rpw import revit, DB
from rpw.db.builtins import BipEnum
from rpw.base import BaseObjectWrapper
from rpw.exceptions import RpwException, RpwWrongStorageType
from rpw.exceptions import RpwParameterNotFound, RpwTypeError
from rpw.utils.logger import logger
from pyrevit.compat import get_elementid_value_func


class ParameterSet(BaseObjectWrapper):
    """
    Allows you to treat an element's parameters as a dictionary.

    This is used internally by Element Wrapper.
    An instance of this class is returned on the ``parameters``
    attribute of wrapped elements.

    >>> element.parameters.all()
    >>> element.parameters['Comments'].value
    >>> element.parameters['Comments'].type

    ParameterSet can also be used for setting values:
    >>> element.parameters['Comments'].value = 'Something'

    >>> parameters = ParameterSet(Element)

    Attributes:
        _revit_object (DB.Element) = Revit Reference

    """

    _revit_object_class = DB.Element

    def __init__(self, element):
        """
        Args:
            element(DB.Element): Element to create ParameterSet
        """
        super(ParameterSet, self).__init__(element)
        self.builtins = _BuiltInParameterSet(self._revit_object)

    def get_value(self, param_name, default_value=None):
        try:
            return self.__getitem__(param_name).value
        except RpwParameterNotFound:
            return default_value

    def __getitem__(self, param_name):
        """ Get's parameter by name.

        Returns:
            :any:`Parameter`: The first parameter found with a matching name (wrapper),

        Raises:
            :class:`RpwParameterNotFound`

        """
        # TODO: Any advantage of using ParameterMap Instead
        parameter = self._revit_object.LookupParameter(param_name)
        # return _Parameter(parameter) if parameter else None
        if not parameter:
            raise RpwParameterNotFound(self._revit_object, param_name)
        return Parameter(parameter)

    def __setitem__(self, param_name, value):
        """ Sets value to element's parameter.
        This is a shorcut to using `parameters['Name'].value = value`

        >>> element.parameters['Height'] = value
        """
        parameter = self.__getitem__(param_name)
        parameter.value = value

    @property
    def all(self):
        """ Returns: Flat list of wrapped parameter elements
        """
        return [Parameter(parameter) for parameter in self._revit_object.Parameters]

    def to_dict(self):
        """ WIP: Returns a Serializable Dictionary """
        return [p.to_dict() for p in self.all]

    def __len__(self):
        return len(self.all)

    def __repr__(self):
        """ Adds data to Base __repr__ to add Parameter List Name """
        return super(ParameterSet, self).__repr__(data={'count': len(self)})


class _BuiltInParameterSet(BaseObjectWrapper):
    """ Built In Parameter Manager

    Usage:

        location_line = element.parameters.builtins['WALL_LOCATION_LINE']

    Note:
        Item Getter can take the BuilInParameter name string, or the Enumeration.
        >>> element.parameters.builtins['WALL_LOCATION_LINE']

        or

        >>>element.parameters.builtins[Revit.DB.BuiltInParameter.WALL_LOCATION_LINE]

    Attributes:
        _revit_object (DB.Element) = Revit Reference

    """

    _revit_object_class = DB.Element

    def __getitem__(self, builtin_enum):
        """ Retrieves Built In Parameter. """
        if isinstance(builtin_enum, str):
            builtin_enum = BipEnum.get(builtin_enum)
        parameter = self._revit_object.get_Parameter(builtin_enum)
        if not parameter:
            raise RpwParameterNotFound(self._revit_object, builtin_enum)
        return Parameter(parameter)

    def __setitem__(self, name, param_value):
        """ Sets value for an element's built in parameter. """
        builtin_parameter = self.__getitem__(name)
        builtin_parameter.value = param_value

    def __repr__(self):
        """ Adds data to Base __repr__ to add Parameter List Name """
        return super(_BuiltInParameterSet, self).__repr__()


class Parameter(BaseObjectWrapper):
    """
    Primarily for internal use by :any:`rpw.db.Element`, but can be used on it's own.

    >>> parameter = Parameter(DB.Parameter)
    >>> parameter.type
    <type: str>
    >>> parameter.value
    'Some String'
    >>> parameter.name
    'Parameter Name'
    >>> parameter.builtin
    Revit.DB.BuiltInParameter.SOME_BUILT_IN_NAME

    Attributes:
        _revit_object (DB.Parameter) = Revit Reference

    Note:

        Parameter Wrapper handles the following types:

        * Autodesk.Revit.DB.StorageType.String
        * Autodesk.Revit.DB.StorageType.Double
        * Autodesk.Revit.DB.StorageType.ElementId
        * Autodesk.Revit.DB.StorageType.Integer
        * Autodesk.Revit.DB.StorageType.None


    """

    _revit_object_class = DB.Parameter
    STORAGE_TYPES = {
                    'String': str,
                    'Double': float,
                    'Integer': int,
                    'ElementId': DB.ElementId,
                    'None': None,
                     }

    def __init__(self, parameter):
        """ Parameter Wrapper Constructor

        Args:
            DB.Parameter: Parameter to be wrapped

        Returns:
            Parameter: Wrapped Parameter Class
        """
        if not isinstance(parameter, DB.Parameter):
            raise RpwTypeError(DB.Parameter, type(parameter))
        super(Parameter, self).__init__(parameter)

    @property
    def type(self):
        """ Returns the Python Type of the Parameter

        >>> element.parameters['Height'].type
        <type: float>

        Returns:
            (``type``): Python Built in type

        """
        storage_type_name = self._revit_object.StorageType.ToString()
        python_type = Parameter.STORAGE_TYPES[storage_type_name]

        return python_type

    @property
    def parameter_type(self):
        parameter_type = self._revit_object.Definition.ParameterType
        return parameter_type

    @property
    def id(self):
        return self._revit_object.Id

    @property
    def value(self):
        """
        Gets Parameter Value:

        >>> desk.parameters['Height'].value
        >>> 3.0

        Sets Parameter Value (must be in Transaction Context):

        >>> desk.parameters['Height'].value = 3

        Returns:
            (``type``): parameter value in python type


        Note:

            `Parameter` value setter automatically handles a few type castings:

            * Storage is ``str`` and value is ``None``; value is converted to ``blank_string``
            * Storage is ``str`` and value is ``any``; value is converted to ``string``
            * Storage is ``ElementId`` and value is ``None``; value is
              converted to ``ElemendId.InvalidElementId``
            * Storage is ``int`` and value is ``float``; value is converted to ``int``
            * Storage is ``float`` and value is ``int``; value is converted to ``float``

        """
        if self.type is str:
            return self._revit_object.AsString()
        if self.type is float:
            return self._revit_object.AsDouble()
        if self.type is DB.ElementId:
            return self._revit_object.AsElementId()
        if self.type is int:
            return self._revit_object.AsInteger()

        raise RpwException('could not get storage type: {}'.format(self.type))

    @value.setter
    def value(self, value):
        if self._revit_object.IsReadOnly:
            definition_name = self._revit_object.Definition.Name
            raise RpwException('Parameter is Read Only: {}'.format(definition_name))

        # Check if value provided matches storage type
        if not isinstance(value, self.type):
            # If not, try to handle
            if self.type is str and value is None:
                value = ''
            if self.type is str and value is not None:
                value = str(value)
            elif self.type is DB.ElementId and value is None:
                value = DB.ElementId.InvalidElementId
            elif isinstance(value, int) and self.type is float:
                value = float(value)
            elif isinstance(value, float) and self.type is int:
                value = int(value)
            elif isinstance(value, bool) and self.type is int:
                value = int(value)
            else:
                raise RpwWrongStorageType(self.type, value)

        param = self._revit_object.Set(value)
        return param

    @property
    def value_string(self):
        """ Ensure Human Readable String Value """
        return self._revit_object.AsValueString() or \
               self._revit_object.AsString()

    def to_dict(self):
        """
        Returns Parameter as a dictionary. Included properties are:

            * name: Parameter.Definition.Name
            * type: Parameter.StorageType.Name
            * value: Uses best parameter method based on StorageType
            * value_string: Parameter.AsValueString
        """
        if not isinstance(self.value, DB.ElementId):
            value = self.value
        else:
            get_elementid_value = get_elementid_value_func()
            value = get_elementid_value(self.value)
        return {
                'name': self.name,
                'type': self.type.__name__,
                'value': value,
                'value_string': self.value_string
                }

    def __bool__(self):
        return bool(self.value)

    def __eq__(self, other):
        """ Equal Parameter Value Comparison """
        return self.value == other

    def __ne__(self, other):
        """ Not Equal Parameter Value Comparison """
        return self.value != other

    def __gt__(self, other):
        """ Greater than Parameter Value Comparison """
        return self.value > other

    def __ge__(self, other):
        """ Greater or Equal Parameter Value Comparison """
        return self.value >= other

    def __lt__(self, other):
        """ Less than Parameter Value Comparison """
        return self.value < other

    def __le__(self, other):
        """ Less or Equal Parameter Value Comparison """
        return self.value <= other

    @property
    def name(self):
        """
        Returns Parameter name

        >>> element.parameters['Comments'].name
        >>> 'Comments'
        """
        return self._revit_object.Definition.Name

    @property
    def builtin(self):
        """ Returns the BuiltInParameter name of Parameter.
        Same as DB.Parameter.Definition.BuiltIn

        Usage:
            >>> element.parameters['Comments'].builtin_name
            Revit.DB.BuiltInParameter.ALL_MODEL_INSTANCE_COMMENTS

        Returns:
            Revit.DB.BuiltInParameter: BuiltInParameter Enumeration Member
        """
        return self._revit_object.Definition.BuiltInParameter

    @property
    def builtin_id(self):
        """ ElementId of BuiltIn

        Usage:
            >>> wall.parameters['Unconnected Height'].builtin_id
            Revit.DB.BuiltInParameter.WALL_USER_HEIGHT_PARAM
        """
        return DB.ElementId(self.builtin)

    def __repr__(self):
        """ Adds data to Base __repr__ to add selection count"""
        return super(Parameter, self).__repr__(data={
                                            'name': self.name,
                                            'value': self.value,
                                            'type': self.type.__name__
                                            })



class BestFitMappingAttribute(Attribute,_Attribute):
 """
 Controls whether Unicode characters are converted to the closest matching ANSI characters.

 

 BestFitMappingAttribute(BestFitMapping: bool)
 """
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 @staticmethod
 def __new__(self,BestFitMapping):
  """ __new__(cls: type,BestFitMapping: bool) """
  pass
 BestFitMapping=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the best-fit mapping behavior when converting Unicode characters to ANSI characters.



Get: BestFitMapping(self: BestFitMappingAttribute) -> bool



"""


 ThrowOnUnmappableChar=None




class AnimationClock(Clock):
 """ Maintains the run-time state of an System.Windows.Media.Animation.AnimationTimeline and processes its output values. """
 def GetCurrentValue(self,defaultOriginValue,defaultDestinationValue):
  """
  GetCurrentValue(self: AnimationClock,defaultOriginValue: object,defaultDestinationValue: object) -> object

  

   Gets the current output value of the System.Windows.Media.Animation.AnimationClock.

  

   defaultOriginValue: The origin value provided to the clock if its animation does not have its own start value. If 

    this clock is the first in a composition chain it will be the base value of the property being 

    animated; otherwise it will be the value returned by the previous clock in the chain

  

   defaultDestinationValue: The destination value provided to the clock if its animation does not have its own destination 

    value. If this clock is the first in a composition chain it will be the base value of the 

    property being animated; otherwise it will be the value returned by the previous clock in the 

    chain

  

   Returns: The current value of this System.Windows.Media.Animation.AnimationClock.
  """
  pass
 @staticmethod
 def __new__(self,*args): #cannot find CLR constructor
  """ __new__(cls: type,animation: AnimationTimeline) """
  pass
 CurrentGlobalTime=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the current global time,as established by the WPF timing system.



"""

 Timeline=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the System.Windows.Media.Animation.AnimationTimeline that describes this clock's behavior.



Get: Timeline(self: AnimationClock) -> AnimationTimeline



"""





"""

rlcompleter from native library. Modified for Dynamo Compatibility

"""


__all__ = ["Completer"]

try:
    import __builtin__
    import keyword
    builtin_dir = __builtin__.__dict__
except:
    builtin_dir = {}
    keyword = []
    pass

class Completer:
    def __init__(self, namespace):
        """Create a new completer for the command line.

        Completer([namespace]) -> completer instance.

        If unspecified, the default namespace where completions are performed
        is __main__ (technically, __main__.__dict__). Namespaces should be
        given as dictionaries.

        Completer instances should be used as the completion mechanism of
        readline via the set_completer() call:

        readline.set_completer(Completer(my_namespace).complete)
        """

        if namespace and not isinstance(namespace, dict):
            raise TypeError,'namespace must be a dictionary'

        self.namespace = namespace

    def complete(self, text, state):
        """Return the next possible completion for 'text'.

        This is called successively with state == 0, 1, 2, ... until it
        returns None.  The completion should begin with 'text'.

        """
        if state == 0:
            if "." in text:
                self.matches = self.attr_matches(text)
            else:
                self.matches = self.global_matches(text)
        try:
            return self.matches[state]
        except IndexError:
            return None

    def _callable_postfix(self, val, word):
        if hasattr(val, '__call__'):
            word = word + "("
        return word

    def global_matches(self, text):
        """Compute matches when text is a simple name.

        Return a list of all keywords, built-in functions and names currently
        defined in self.namespace that match.

        """
        matches = []
        seen = {"__builtins__"}
        n = len(text)
        if keyword:
            for word in keyword.kwlist:
                if word[:n] == text:
                    seen.add(word)
                    matches.append(word)
        for nspace in [self.namespace, builtin_dir]:
            for word, val in nspace.items():
                if word[:n] == text and word not in seen:
                    seen.add(word)
                    matches.append(self._callable_postfix(val, word))
        return matches

    def attr_matches(self, text):
        """Compute matches when text contains a dot.

        Assuming the text is of the form NAME.NAME....[NAME], and is
        evaluable in self.namespace, it will be evaluated and its attributes
        (as revealed by dir()) are used as possible completions.  (For class
        instances, class members are also considered.)

        WARNING: this can still invoke arbitrary C code, if an object
        with a __getattr__ hook is evaluated.

        """
        import re
        m = re.match(r"(\w+(\.\w+)*)\.(\w*)", text)
        if not m:
            return []
        expr, attr = m.group(1, 3)
        try:
            thisobject = eval(expr, self.namespace)
        except Exception:
            return []

        # get the content of the object, except __builtins__
        words = set(dir(thisobject))
        words.discard("__builtins__")

        if hasattr(thisobject, '__class__'):
            words.add('__class__')
            words.update(get_class_members(thisobject.__class__))
        matches = []
        n = len(attr)
        for word in words:
            if word[:n] == attr:
                try:
                    val = getattr(thisobject, word)
                except Exception:
                    continue  # Exclude properties that are not set
                word = self._callable_postfix(val, "%s.%s" % (expr, word))
                matches.append(word)
        matches.sort()
        return matches

def get_class_members(klass):
    ret = dir(klass)
    if hasattr(klass,'__bases__'):
        for base in klass.__bases__:
            ret = ret + get_class_members(base)
    return ret



"""Event telemetry management."""
#pylint: disable=invalid-name,broad-except
from pyrevit import HOST_APP
from pyrevit.coreutils import logger



mlogger = logger.get_logger(__name__)


# event telemetry configurations are controlled by a binary flag
# flag is assumed to be 16 bytes long (128 bits)
# although both python and C# implementation use flexible integers
# the flags are sorted by the index of EventType enum values
ALL_EVENTS = 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF

# flag for a suggested configuration for event telementry
SUGGESTED_EVENTS = 0x1d17e7f7ff59


def register_event_telemetry(handler, flags):
    """Registers application event telemetry handlers based on given flags.

    Args:
        handler (EventTelemetry): event telemetry handler
        flags (int): event flags
    """
    try:
        handler.RegisterEventTelemetry(HOST_APP.uiapp, flags)
    except Exception as ex:
        mlogger.debug(
            "Error registering event telementry with flags: %s | %s",
            str(flags), ex)


def unregister_event_telemetry(handler, flags):
    """Unregisters application event telemetry handlers based on given flags.

    Args:
        handler (EventTelemetry): event telemetry handler
        flags (int): event flags
    """
    try:
        handler.UnRegisterEventTelemetry(HOST_APP.uiapp, flags)
    except Exception as ex:
        mlogger.debug(
            "Error unregistering event telementry with flags: %s | %s",
            str(flags), ex)


def unregister_all_event_telemetries(handler):
    """Unregisters all available application event telemetry handlers."""
    unregister_event_telemetry(handler, ALL_EVENTS)



class DataFormats(object):
 """ Provides static,predefined System.Windows.Forms.Clipboard format names. Use them to identify the format of data that you store in an System.Windows.Forms.IDataObject. """
 @staticmethod
 def GetFormat(*__args):
  """
  GetFormat(id: int) -> Format

  

   Returns a System.Windows.Forms.DataFormats.Format with the Windows Clipboard numeric ID and name 

    for the specified ID.

  

  

   id: The format ID.

   Returns: A System.Windows.Forms.DataFormats.Format that has the Windows Clipboard numeric ID and the name 

    of the format.

  

  GetFormat(format: str) -> Format

  

   Returns a System.Windows.Forms.DataFormats.Format with the Windows Clipboard numeric ID and name 

    for the specified format.

  

  

   format: The format name.

   Returns: A System.Windows.Forms.DataFormats.Format that has the Windows Clipboard numeric ID and the name 

    of the format.
  """
  pass
 Bitmap='Bitmap'
 CommaSeparatedValue='Csv'
 Dib='DeviceIndependentBitmap'
 Dif='DataInterchangeFormat'
 EnhancedMetafile='EnhancedMetafile'
 FileDrop='FileDrop'
 Format=None
 Html='HTML Format'
 Locale='Locale'
 MetafilePict='MetaFilePict'
 OemText='OEMText'
 Palette='Palette'
 PenData='PenData'
 Riff='RiffAudio'
 Rtf='Rich Text Format'
 Serializable='WindowsForms10PersistentObject'
 StringFormat='System.String'
 SymbolicLink='SymbolicLink'
 Text='Text'
 Tiff='TaggedImageFileFormat'
 UnicodeText='UnicodeText'
 WaveAudio='WaveAudio'




class DataGridViewRowsRemovedEventArgs(EventArgs):
 """
 Provides data for the System.Windows.Forms.DataGridView.RowsRemoved event.

 

 DataGridViewRowsRemovedEventArgs(rowIndex: int,rowCount: int)
 """
 @staticmethod
 def __new__(self,rowIndex,rowCount):
  """ __new__(cls: type,rowIndex: int,rowCount: int) """
  pass
 RowCount=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the number of rows that were deleted.



Get: RowCount(self: DataGridViewRowsRemovedEventArgs) -> int



"""

 RowIndex=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the zero-based index of the row deleted,or the first deleted row if multiple rows were deleted.



Get: RowIndex(self: DataGridViewRowsRemovedEventArgs) -> int



"""





class Int32(object):
 """ Represents a 32-bit signed integer. """
 def bit_length(self,*args):
  """ bit_length(value: int) -> int """
  pass
 def conjugate(self,*args):
  """ conjugate(x: int) -> int """
  pass
 def __abs__(self,*args):
  """ x.__abs__() <==> abs(x) """
  pass
 def __add__(self,*args):
  """ x.__add__(y) <==> x+y """
  pass
 def __and__(self,*args):
  """ __and__(x: int,y: int) -> int """
  pass
 def __cmp__(self,*args):
  """ x.__cmp__(y) <==> cmp(x,y) """
  pass
 def __coerce__(self,*args):
  """ __coerce__(x: int,o: object) -> object """
  pass
 def __divmod__(self,*args):
  """
  __divmod__(x: int,y: object) -> object

  __divmod__(x: int,y: int) -> tuple
  """
  pass
 def __div__(self,*args):
  """ x.__div__(y) <==> x/y """
  pass
 def __float__(self,*args):
  """ __float__(self: int) -> float """
  pass
 def __floordiv__(self,*args):
  """ x.__floordiv__(y) <==> x//y """
  pass
 def __getnewargs__(self,*args):
  """ __getnewargs__(self: int) -> object """
  pass
 def __hex__(self,*args):
  """ __hex__(x: int) -> str """
  pass
 def __index__(self,*args):
  """ __index__(x: int) -> int """
  pass
 def __int__(self,*args):
  """ __int__(self: int) -> int """
  pass
 def __invert__(self,*args):
  """ __invert__(x: int) -> int """
  pass
 def __long__(self,*args):
  """ __long__(self: int) -> long """
  pass
 def __lshift__(self,*args):
  """ x.__rshift__(y) <==> x<<yx.__rshift__(y) <==> x<<y """
  pass
 def __mod__(self,*args):
  """ x.__mod__(y) <==> x%y """
  pass
 def __mul__(self,*args):
  """ x.__mul__(y) <==> x*y """
  pass
 def __neg__(self,*args):
  """ x.__neg__() <==> -x """
  pass
 @staticmethod
 def __new__(self,*args): #cannot find CLR constructor
  """
  __new__(cls: type,s: IList[Byte]) -> object

  __new__(cls: type,x: object) -> object

  __new__(cls: type) -> object

  __new__(o: object) -> object

  __new__(cls: type,o: Extensible[float]) -> object

  __new__(cls: type,s: str,base: int) -> object
  """
  pass
 def __nonzero__(self,*args):
  """ __nonzero__(x: int) -> bool """
  pass
 def __oct__(self,*args):
  """ __oct__(x: int) -> str """
  pass
 def __or__(self,*args):
  """ __or__(x: int,y: int) -> int """
  pass
 def __pos__(self,*args):
  """ __pos__(x: int) -> int """
  pass
 def __pow__(self,*args):
  """ x.__pow__(y[,z]) <==> pow(x,y[,z])x.__pow__(y[,z]) <==> pow(x,y[,z])x.__pow__(y[,z]) <==> pow(x,y[,z])x.__pow__(y[,z]) <==> pow(x,y[,z]) """
  pass
 def __radd__(self,*args):
  """ __radd__(x: int,y: int) -> object """
  pass
 def __rand__(self,*args):
  """ __rand__(x: int,y: int) -> int """
  pass
 def __rdivmod__(self,*args):
  """ __rdivmod__(x: int,y: int) -> object """
  pass
 def __rdiv__(self,*args):
  """ __rdiv__(x: int,y: int) -> object """
  pass
 def __rfloordiv__(self,*args):
  """ __rfloordiv__(x: int,y: int) -> object """
  pass
 def __rlshift__(self,*args):
  """ __rlshift__(x: int,y: int) -> object """
  pass
 def __rmod__(self,*args):
  """ __rmod__(x: int,y: int) -> int """
  pass
 def __rmul__(self,*args):
  """ __rmul__(x: int,y: int) -> object """
  pass
 def __ror__(self,*args):
  """ __ror__(x: int,y: int) -> int """
  pass
 def __rpow__(self,*args):
  """
  __rpow__(x: int,power: int,qmod: Nullable[int]) -> object

  __rpow__(x: int,power: int) -> object

  __rpow__(x: int,power: long,qmod: long) -> object

  __rpow__(x: int,power: float,qmod: float) -> object
  """
  pass
 def __rrshift__(self,*args):
  """ __rrshift__(x: int,y: int) -> int """
  pass
 def __rshift__(self,*args):
  """ x.__rshift__(y) <==> x>>yx.__rshift__(y) <==> x>>y """
  pass
 def __rsub__(self,*args):
  """ __rsub__(x: int,y: int) -> object """
  pass
 def __rtruediv__(self,*args):
  """ __rtruediv__(x: int,y: int) -> float """
  pass
 def __rxor__(self,*args):
  """ __rxor__(x: int,y: int) -> int """
  pass
 def __sub__(self,*args):
  """ x.__sub__(y) <==> x-y """
  pass
 def __truediv__(self,*args):
  """ x.__truediv__(y) <==> x/y """
  pass
 def __trunc__(self,*args):
  """ __trunc__(x: int) -> int """
  pass
 def __xor__(self,*args):
  """ __xor__(x: int,y: int) -> int """
  pass
 denominator=None
 imag=None
 numerator=None
 real=None




class QualifiedAce(KnownAce):
 """ Represents an Access Control Entry (ACE) that contains a qualifier. The qualifier,represented by an System.Security.AccessControl.AceQualifier object,specifies whether the ACE allows access,denies access,causes system audits,or causes system alarms. The System.Security.AccessControl.QualifiedAce class is the abstract base class for the System.Security.AccessControl.CommonAce and System.Security.AccessControl.ObjectAce classes. """
 def GetOpaque(self):
  """
  GetOpaque(self: QualifiedAce) -> Array[Byte]

  

   Returns the opaque callback data associated with this System.Security.AccessControl.QualifiedAce 

    object.

  

   Returns: An array of byte values that represents the opaque callback data associated with this 

    System.Security.AccessControl.QualifiedAce object.
  """
  pass
 def SetOpaque(self,opaque):
  """
  SetOpaque(self: QualifiedAce,opaque: Array[Byte])

   Sets the opaque callback data associated with this System.Security.AccessControl.QualifiedAce 

    object.

  

  

   opaque: An array of byte values that represents the opaque callback data for this 

    System.Security.AccessControl.QualifiedAce object.
  """
  pass
 AceQualifier=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value that specifies whether the ACE allows access,denies access,causes system audits,or causes system alarms.



Get: AceQualifier(self: QualifiedAce) -> AceQualifier



"""

 IsCallback=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Specifies whether this System.Security.AccessControl.QualifiedAce object contains callback data.



Get: IsCallback(self: QualifiedAce) -> bool



"""

 OpaqueLength=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the length of the opaque callback data associated with this System.Security.AccessControl.QualifiedAce object. This property is valid only for callback Access Control Entries (ACEs).



Get: OpaqueLength(self: QualifiedAce) -> int



"""


